/*
 * blockserver_io.c --- L4Re block server I/O manager.
 *
 * Copyright (C) 2019, 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */

#include <l4/re/env.h>

#include <stdio.h>
#include <stdlib.h>

#include <ext2fs/ext2fs.h>
#include <fsclient/client.h>



/* Specific data for interacting with block devices. */

struct blockserver_private_data
{
  int magic;
  int flags;
  file_t *file;
};

io_manager blockserver_io_manager; /* defined below */

const char *blockserver_default_cap = "server";



/* Allocate and initialise a channel for the indicated block device. */

static errcode_t blockserver_open_channel(const char *name,
                                          file_t *file, int flags,
                                          io_channel *channel,
                                          io_manager io_mgr)
{
  io_channel io = NULL;
  struct blockserver_private_data *data = NULL;
  errcode_t retval;

  /* Allocate memory for the channel. */

  retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
  if (retval)
    goto free_io;

  /* Initialise the channel. */

  memset(io, 0, sizeof(struct struct_io_channel));

  io->magic = EXT2_ET_MAGIC_IO_CHANNEL;
  io->manager = io_mgr;
  io->block_size = 1024;
  io->read_error = 0;
  io->write_error = 0;
  io->refcount = 1;

  /* Allocate memory for the name and store it. */

  retval = ext2fs_get_mem(strlen(name)+1, &io->name);
  if (retval)
    goto free_io;

  strcpy(io->name, name);

  /* Allocate memory for the private data. */

  retval = ext2fs_get_mem(sizeof(struct blockserver_private_data), &data);
  if (retval)
    goto free_data;

  /* Initialise the private data. */

  io->private_data = data;

  memset(data, 0, sizeof(struct blockserver_private_data));

  data->magic = EXT2_ET_MAGIC_UNIX_IO_CHANNEL;
  data->flags = flags;
  data->file = file;

  /* Return the allocated channel. */

  *channel = io;
  return 0;

  /* Deallocation upon failure. */

free_data:
  if (data)
    ext2fs_free_mem(&data);

free_io:
  if (io)
  {
    if (io->name)
      ext2fs_free_mem(&io->name);
    ext2fs_free_mem(&io);
  }

  return retval;
}



/* Open a block device using the named capability. */

static errcode_t blockserver_open(const char *name, int flags, io_channel *channel)
{
  l4_cap_idx_t server = l4re_env_get_cap(blockserver_default_cap);

  /* Open a device via the named capability. */

  file_t *file = client_open_using(name, flags, server);

  /* NOTE: May want a more appropriate error code. */

  if (file == NULL)
    return EXT2_ET_BAD_DEVICE_NAME;

  return blockserver_open_channel(name, file, flags, channel, blockserver_io_manager);
}



/* Close the block device associated with the given channel. */

static errcode_t blockserver_close(io_channel channel)
{
  struct blockserver_private_data *data;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct blockserver_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

  client_close(data->file);
  return 0;
}

/* Update the block size in use. */

static errcode_t blockserver_set_blksize(io_channel channel, int blksize)
{
  struct blockserver_private_data *data;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct blockserver_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

  if (channel->block_size != blksize)
    channel->block_size = blksize;

  return 0;
}

/* Flush unwritten data. Without a cache, this does nothing. */

static errcode_t blockserver_flush(io_channel channel)
{
  (void) channel;
  return 0;
}

static errcode_t blockserver_set_option(io_channel channel, const char *option, const char *arg)
{
  (void) channel; (void) option; (void) arg;
  return 0;
}

static errcode_t blockserver_get_stats(io_channel channel, io_stats *stats)
{
  (void) channel; (void) stats;
  return 0;
}

static errcode_t blockserver_read_blk64(io_channel channel, unsigned long long block, int count, void *buf)
{
  /* Negative count values indicate a precise amount to read. */

  unsigned long long to_read = count < 0 ? -count : count * channel->block_size;
  offset_t read;
  struct blockserver_private_data *data;
  errcode_t retval = 0;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct blockserver_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

  /* Perform the read at the given location. */

  if (client_seek(data->file, block * channel->block_size, SEEK_SET) != block * channel->block_size)
  {
    read = 0;
    retval = EXT2_ET_LLSEEK_FAILED;
  }
  else
  {
    read = client_read(data->file, buf, to_read);

    /* Check for a short read. */

    if (read < to_read)
      retval = EXT2_ET_SHORT_READ;
  }

  /* Handle read errors generally. */

  if (retval)
    if (channel->read_error)
      retval = (channel->read_error)(channel, block, count, buf,
                                     to_read, read, retval);

  return retval;
}

static errcode_t blockserver_write_blk64(io_channel channel, unsigned long long block, int count, const void *buf)
{
  /* Negative count values indicate a precise amount to write. */

  unsigned long long to_write = count < 0 ? -count : count * channel->block_size;
  offset_t written;
  struct blockserver_private_data *data;
  errcode_t retval = 0;

  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
  data = (struct blockserver_private_data *) channel->private_data;
  EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

  /* Perform the write at the given location. */

  if (client_seek(data->file, block * channel->block_size, SEEK_SET) != block * channel->block_size)
  {
    retval = EXT2_ET_LLSEEK_FAILED;
    written = 0;
  }
  else
  {
    written = client_write(data->file, buf, to_write);

    /* Check for a short write. */

    if (written < to_write)
      retval = EXT2_ET_SHORT_WRITE;
  }

  /* Handle write errors generally. */

  if (retval)
    if (channel->write_error)
      retval = (channel->write_error)(channel, block, count, buf,
                                      to_write, written, retval);

  return retval;
}

static errcode_t blockserver_read_blk(io_channel channel, unsigned long block, int count, void *buf)
{
  return blockserver_read_blk64(channel, block, count, buf);
}

static errcode_t blockserver_write_blk(io_channel channel, unsigned long block, int count, const void *buf)
{
  return blockserver_write_blk64(channel, block, count, buf);
}

/* Unsupported operations are omitted: see io_manager.c for optional operation
   details. */

static struct struct_io_manager struct_blockserver_manager = {
  .magic            = EXT2_ET_MAGIC_IO_MANAGER,
  .name             = "L4Re blockserver I/O Manager",
  .open             = blockserver_open,
  .close            = blockserver_close,
  .set_blksize      = blockserver_set_blksize,
  .read_blk         = blockserver_read_blk,
  .write_blk        = blockserver_write_blk,
  .flush            = blockserver_flush,
  .write_byte       = 0,
  .set_option       = blockserver_set_option,
  .get_stats        = blockserver_get_stats,
  .read_blk64       = blockserver_read_blk64,
  .write_blk64      = blockserver_write_blk64,
  .discard          = 0,
  .cache_readahead  = 0,
  .zeroout          = 0,
};

io_manager blockserver_io_manager = &struct_blockserver_manager;
